classdef TestPgmresSolver < matlab.unittest.TestCase
    
properties (TestParameter)
    p = struct('low', 1);
    variant = struct('loAdditiveSchwarzPgmres', "additiveSchwarzLowOrder");
end

methods (Test)
    function firstStepDecreasesNorm(testCase, p, variant)
        [mesh, fes, blf, lf] = setupProblem(testCase, p);
        s = variant(1);
        solver = PgmresSolverSetup.choose(fes, blf, s);
        
        [xstar, A, symA, F] = assembleData(testCase, blf, lf, fes);
        solver.setupSystemMatrix(A, symA);
        solver.setupRhs(F);

        normBefore = sqrt((xstar - solver.x)' * symA * (xstar - solver.x));
        solver.solve();
        normAfterwards = sqrt((xstar - solver.x)' * symA * (xstar - solver.x));

        u = FeFunction(fes);
        u.setFreeData(solver.x);
        
        testCase.verifyGreaterThan(normBefore, normAfterwards);
    end

    function laterStepDecreasesNorm(testCase, p, variant)
        [mesh, fes, blf, lf] = setupProblem(testCase, p);
        s = variant(1);
        solver = PgmresSolverSetup.choose(fes, blf, s);
        
        for k = 1:3
            [xstar, A, symA, F] = assembleData(testCase, blf, lf, fes);
            solver.setupSystemMatrix(A, symA);
            if k < 3, mesh.refineUniform(); end
        end
        
        solver.setupRhs(F);
        normBefore = sqrt((xstar - solver.x)' * symA * (xstar - solver.x));
        solver.solve();
        normAfterwards = sqrt((xstar - solver.x)' * symA * (xstar - solver.x));
        
        testCase.verifyGreaterThan(normBefore, normAfterwards);
    end
end

methods (Access=private)
    function [mesh, fes, blf, lf] = setupProblem(~, p)
        mesh = Mesh.loadFromGeometry('Lshape');
        mesh.refineUniform(5);
        fes = FeSpace(mesh, HigherOrderH1Fe(p), 'dirichlet', ':');
        blf = BilinearForm();
        lf = LinearForm();
        blf.a = Constant(mesh, 1);
        blf.b = Constant(mesh, [10;10]);
        lf.f = Constant(mesh, 1);
    end
    
    function [xstar, A, symA, F] = assembleData(~, blf, lf, fes)
        A = assemble(blf, fes);
        principalPart = BilinearForm();
        principalPart.a = blf.a;
        symA = assemble(principalPart, fes);
        F = assemble(lf, fes);
        freeDofs = getFreeDofs(fes);
        A = A(freeDofs,freeDofs);
        symA = symA(freeDofs, freeDofs);
        F = F(freeDofs);
        xstar = A \ F;
    end
end
    
    
end